import copy
from operator import itemgetter

from twisted.internet.defer import inlineCallbacks, returnValue

from ensemble.environment.errors import EnvironmentsConfigError

from .bootstrap import Bootstrap
from .connect import ZookeeperConnect
from .findzookeepers import find_zookeepers
from .state import SaveState, LoadState
from .utils import get_user_authorized_keys


class MachineProviderBase(object):
    """Base class supplying common functionality for MachineProviders.

    To write a working subclass, you will need to override the following
    methods:

        * get_file_storage
        * start_machine
        * get_machines
        * shutdown_machines
        * open_port
        * close_port
        * get_opened_ports

    You may want to override the following methods, but you should be careful
    to call MachineProviderBase's implementation (or be very sure you don't
    need to:

        * __init__
        * get_serialization_data

    You probably shouldn't override anything else.
    """

    def __init__(self, environment_name, config):
        if ("authorized-keys-path" in config and
            "authorized-keys" in config):
            raise EnvironmentsConfigError(
                "Environment config cannot define both authorized-keys "
                "and authorized-keys-path. Pick one!")

        self.environment_name = environment_name
        self.config = config

    def get_serialization_data(self):
        """Get provider configuration suitable for serialization."""
        data = copy.deepcopy(self.config)
        data["authorized-keys"] = get_user_authorized_keys(data)
        # Not relevant, on a remote system.
        data.pop("authorized-keys-path", None)
        return data

    #================================================================
    # Subclasses need to implement their own versions of everything
    # in the following block

    def get_file_storage(self):
        """Retrieve the provider FileStorage abstraction."""
        raise NotImplementedError()

    def start_machine(self, machine_data, master=False):
        """Start a machine in the provider.

        @param machine_data: a dictionary of data to pass along to the newly
           launched machine.

        @param master: if True, machine will initialize the ensemble admin
            and run a provisioning agent.
        """
        raise NotImplementedError()

    def get_machines(self, instance_ids=()):
        """List machines running in the provider.

        @param instance_ids: ids of instances you want to get. Leave empty
            to list all machines owned by this provider.

        @return: a list of provider-specific ProviderMachines.

        @raise: MachinesNotFound
        """
        raise NotImplementedError()

    def shutdown_machines(self, machines):
        """Terminate machines associated with this provider.

        @param machines: list of machines to shut down
        """
        raise NotImplementedError()

    def open_port(self, machine, machine_id, port, protocol="tcp"):
        """Authorizes `port` using `protocol` for `machine`."""
        raise NotImplementedError()

    def close_port(self, machine, machine_id, port, protocol="tcp"):
        """Revokes `port` using `protocol` for `machine`."""
        raise NotImplementedError()

    def get_opened_ports(self, machine, machine_id):
        """Returns a set of open (port, protocol) pairs for `machine`."""
        raise NotImplementedError()

    #================================================================
    # Subclasses will not generally need to override the methods in
    # this block

    def get_zookeeper_machines(self):
        """Find running zookeeper instances.

        @return: the first valid instance found as a single element list.

        @raise: `EnvironmentNotFound`
        """
        return find_zookeepers(self)

    def connect(self, share=False):
        """Attempt to connect to a running zookeeper node.

        @param share: where feasible, attempt to share a connection with other
            clients

        @return: an open `txzookeeper.client.ZookeeperClient`

        @raise: `EnvironmentNotFound` when no zookeepers exist
        @raise: `EnvironmentPending` when zookeepers exist but connect fails
        """
        return ZookeeperConnect(self).run(share=share)

    def bootstrap(self):
        """Bootstrap an ensemble server in the provider."""
        return Bootstrap(self).run()

    def get_machine(self, instance_id):
        """Retrieve a provider machine by instance id.

        @param instance_id: instance_id of the `ProviderMachine` you want.

        @return: a single `ProviderMachine`

        @raise: `MachinesNotFound`
        """
        d = self.get_machines([instance_id])
        d.addCallback(itemgetter(0))
        return d

    def shutdown_machine(self, machine):
        """Terminate one machine associated with this provider.

        @param machine: ProviderMachine to shut down.

        @raise: MachinesNotFound
        """
        d = self.shutdown_machines([machine])
        d.addCallback(itemgetter(0))
        return d

    @inlineCallbacks
    def destroy_environment(self):
        """Clear ensemble state and terminate all associated machines"""
        try:
            yield self.save_state({})
        finally:
            live_machines = yield self.get_machines()
            killed_machines = yield self.shutdown_machines(live_machines)
            returnValue(killed_machines)

    def save_state(self, state):
        """Save state to the provider.

        @param state: a dictionary to persist
        """
        return SaveState(self).run(state)

    def load_state(self):
        """Load state from the provider.

        @return: a previously-persisted dictionary, or False
        """
        return LoadState(self).run()
